import io
import itertools

import logging
import requests
from requests.auth import HTTPBasicAuth

from ._types import Product

logger = logging.getLogger(__name__)


class ShopifyClient:

    def __init__(self, *, store_name, api_key, password):
        self.store_name = store_name
        self.api_key = api_key
        self.password = password

    def _make_base_url(self):
        return 'https://{}.myshopify.com/'.format(self.store_name)

    def _make_url(self, path):
        return '{}{}'.format(self._make_base_url(), path)

    def _make_auth(self):
        return HTTPBasicAuth(self.api_key, self.password)

    def request(self, method, path, *args, **kwargs):
        url = self._make_url(path)
        kwargs.setdefault('auth', self._make_auth())
        resp = requests.request(method, url, *args, **kwargs)
        if not resp.ok:
            raise http_error_exception(resp)
        return resp

    def __repr__(self):
        return (
            "ShopifyClient(store_name={}, api_key={})"
            .format(repr(self.store_name), repr(self.api_key)))

    # ================================================================
    # Products API
    # ================================================================

    def get_product_by_id(self, product_id):
        resp = self.request('GET', 'admin/products/{}.json'.format(product_id))
        return resp.json()

    def get_product_by_handle(self, handle):
        products = self.list_all_products({'handle': handle})
        for prod in products:
            return prod
        return None

    def list_all_products(self, filters={}):
        for page in itertools.count(1):
            resp = self.request(
                'GET', 'admin/products.json',
                params={
                    'limit': '250',
                    'page': str(page),
                    **filters,
                })

            products = resp.json()['products']
            if not len(products):
                break  # We reached an end

            for product in products:
                yield Product.from_response(product)

    def create_product(self, product):
        resp = self.request(
            'POST', 'admin/products.json',
            json={'product': product})
        return Product.from_response(resp.json())

    def update_product(self, product_id, updates):

        if 'metafields' in updates:
            # FIXME: need to pass the metafield ID to update
            del updates['metafields']

        resp = self.request(
            'PUT', 'admin/products/{}.json'.format(product_id),
            json={'product': updates})

        return resp.json()

    def delete_product(self, product_id):
        self.request(
            'DELETE', 'admin/products/{}.json'.format(product_id))

    def create_or_update_product(self, product):

        if product.get('id'):
            _product = self.get_product_by_id(product['id'])
            if _product is not None:
                logger.debug('Product found by ID ({}) -> updating'
                             .format(product['id']))
                return self.update_product(product['id'], product)

        handle = product.get('handle')
        if handle:
            _product = self.get_product_by_handle(handle)
            if _product is not None:
                logger.debug('Product found by handle ({}) -> updating'
                             .format(handle))
                return self.update_product(_product['id'], product)

        logger.debug('Creating a new product: {}'.format(repr(product)))
        return self.create_product(product)


class HttpError(Exception):
    pass


def http_error_exception(resp):
    # TODO: exception should be formatted at print time
    # term_width = click.get_terminal_size()[0]
    term_width = 80
    assert not resp.ok
    msg = io.StringIO()

    msg.write(' Request '.center(term_width, '-') + '\n')
    msg.write('{} {}\n'.format(resp.request.method, resp.request.url))
    for key, val in resp.request.headers.items():
        msg.write('{}: {}\n'.format(key, val))
    if resp.request.body:
        msg.write('\n')
        msg.write(format(resp.request.body))
    msg.write('\n\n')

    msg.write(' Response '.center(term_width, '-') + '\n')
    msg.write('{} {}\n'.format(resp.status_code, resp.reason))
    for key, val in resp.request.headers.items():
        msg.write('{}: {}\n'.format(key, val))
    if resp.text:
        msg.write('\n')
        msg.write(resp.text)  # Or resp.content ?
    msg.write('\n')

    msg.write('-' * term_width + '\n')

    return HttpError(msg.getvalue())



import json
import os
import warnings


class ConfigManager:

    def __init__(self, config_home):
        self.config_home = config_home
        if not os.path.exists(self.config_home):
            os.makedirs(self.config_home, exist_ok=True)

    def load_connections(self):
        return self._load_json('connections')

    def save_connections(self, data):
        return self._save_json('connections', data)

    def create_connection(self, name, config):
        data = self.load_connections()
        if name in data:
            raise ValueError(
                'Connection {} exists! Not overwriting.'
                .format(name))
        data[name] = config
        self.save_connections(data)

    def get_connection(self, name):
        data = self.load_connections()
        try:
            return data[name]
        except KeyError:
            pass
        raise KeyError('Connection does not exist: {}'.format(name))

    def update_connection(self, name, config):
        data = self.load_connections()
        if name not in data:
            warnings.warn(
                'Connection does not exist: {}. Creating a new one.'
                .format(name))
        data[name] = config
        self.save_connections(data)

    def delete_connection(self, name):
        data = self.load_connections()
        data.pop(name, None)
        self.save_connections(data)

    def _load_json(self, name):
        filename = self._get_json_file_name(name)
        try:
            with open(filename, 'r') as fp:
                return json.load(fp)
        except FileNotFoundError:
            return {}

    def _save_json(self, name, data):
        filename = self._get_json_file_name(name)
        with open(filename, 'w') as fp:
            json.dump(data, fp)
        # TODO: only set file mode on creation?
        os.chmod(filename, mode=0o600)

    def _get_json_file_name(self, name):
        return os.path.join(self.config_home, '{}.json'.format(name))



import json
import logging
import os
from functools import partial, update_wrapper

import click

import nicelog

from ._config import ConfigManager
from ._output import highlight_json, print_object
from ._utils import string_limit_visual_length
from .client import ShopifyClient

logger = logging.getLogger(__name__)

_home = os.path.expanduser('~')
xdg_config_home = (
    os.environ.get('XDG_CONFIG_HOME') or
    os.path.join(_home, '.config'))


cfg = ConfigManager(os.path.join(xdg_config_home, 'shopify'))


def with_shopify_client(f):
    """Decorator to pass client from context
    """

    @click.pass_context
    def new_func(ctx, *args, **kwargs):
        conn_name = ctx.obj['CONNECTION_NAME']
        client = _get_store_client(conn_name) if conn_name else None
        logger.debug('Shopify client: {}'.format(client))
        return ctx.invoke(f, client, *args, **kwargs)

    return update_wrapper(new_func, f)


def _get_store_client(connection):
    try:
        credentials = cfg.get_connection(connection)
    except KeyError:
        logger.warning('Connection not found: {}'.format(connection))
        return None

    client = ShopifyClient(
        store_name=credentials['store'],
        api_key=credentials['api_key'],
        password=credentials['password'])
    return client


def with_output_serializer(f):

    @click.pass_context
    def new_func(ctx, *args, **kwargs):
        fmtname = ctx.obj['OUTPUT_FORMAT']
        serializer = partial(print_object, fmt=fmtname)
        return ctx.invoke(f, *args, **kwargs, print_object=serializer)

    return update_wrapper(new_func, f)


def autocomplete_connection_names(ctx, args, incomplete):
    connections = cfg.load_connections()
    return sorted(
        (name, '{0[store]} {0[api_key]}'.format(params))
        for name, params in connections.items()
        if incomplete in name)


@click.group()
@click.option('--debug/--no-debug', default=False)
@click.option('conn_name', '--connection', '-c', default='default',
              autocompletion=autocomplete_connection_names)
@click.option('output_fmt', '--output-format', '-f', default=None)
@click.pass_context
def main(ctx, debug, conn_name, output_fmt):

    nicelog.setup_logging()
    log_level = logging.DEBUG if debug else logging.INFO
    logging.getLogger().setLevel(log_level)

    ctx.ensure_object(dict)
    ctx.obj['DEBUG'] = debug
    ctx.obj['CONNECTION_NAME'] = conn_name
    ctx.obj['OUTPUT_FORMAT'] = output_fmt


@main.group('connection')
def grp_connection():
    """Manage connections to Shopify stores"""
    pass


def store_name_info(ctx, param, value):
    print(
        'Head to https://{}.myshopify.com/admin/apps/private if you need '
        'to create a new API key.'
        .format(value))
    return value


@grp_connection.command('create')
@click.option('--name', '-n', prompt='Name this connection',
              help='Connection name, used to refer to this connection',
              default='default')
@click.option('--store', '-s', prompt='Store name', callback=store_name_info,
              help='Shopify store name')
@click.option('--api-key', prompt='API key', help='Shopify API key')
@click.option('--password', prompt='API key password', hide_input=True,
              help='Shopify API password')
def cmd_connection_create(name, store, api_key, password):
    """Create a new connection"""
    cfg.create_connection(name, {
        'store': store,
        'api_key': api_key,
        'password': password,
    })


@grp_connection.command('list')
def cmd_connection_list():
    """List available connections"""
    connections = cfg.load_connections()
    for name, params in sorted(connections.items()):
        click.echo(
            '\x1b[1;32m{name}:\x1b[0m \x1b[1m{p[store]}\x1b[0m {p[api_key]}'
            .format(name=name, p=params))


@grp_connection.command('delete')
@click.argument('name')
def cmd_connection_delete(name):
    """Delete a connection"""
    cfg.delete_connection(name)


@main.group('products')
def grp_products():
    """Manage products"""
    pass


@grp_products.command('get')
@click.argument('product_id')
@with_shopify_client
def cmd_products_get(client, product_id):
    product = client.get_product_by_id(product_id)
    print_object(product)


@grp_products.command('list')
@with_shopify_client
@with_output_serializer
def cmd_products_list(client, print_object):
    products = client.list_all_products()
    print_object(products)


@grp_products.command('export')
@with_shopify_client
@with_output_serializer
def cmd_products_export(client, print_object):
    products = client.list_all_products()
    print_object({'products': products})


@main.command('import')
@click.argument('input_file', type=click.File('rb'), default='-')
@click.option('--info', default=False, is_flag=True,
              help='Only show information about the data file')
@with_shopify_client
@with_output_serializer
def cmd_import(client, input_file, info, print_object):
    """Import data to a store"""

    data = json.load(input_file)

    if not isinstance(data, dict):
        raise TypeError('Invalid data file (expected JSON object)')

    if 'products' in data:
        click.echo('\x1b[1;31mProducts:\x1b[0m')
        print_objects_summary(data['products'])

        if not info:
            for product in data['products']:
                logger.debug('Importing product: %s', repr(product))
                client.create_or_update_product(product)


def print_objects_summary(objects):
    width, height = click.get_terminal_size()

    indent = '    - '
    ellipsis = ' ...'
    max_width = width - len(indent)

    for item in objects:
        text = json.dumps(item)

        if len(text) > max_width:
            hilite = highlight_json(text)
            maxlen = max_width - len(ellipsis) - 1
            text = string_limit_visual_length(hilite, maxlen)

        click.echo('{}{}{}'.format(indent, text, ellipsis))


if __name__ == '__main__':
    main()



import re
import io


def string_limit_visual_length(text, maxlen=None):
    """Limit a string to a *visual* length.

    Will cut string to a specified maximum length, ignoring ANSI
    escape codes in determining the length.

    This allows trimming coloured strings to terminal width.
    """

    stream = io.StringIO()

    TOKEN_RE = re.compile(
        r'(?P<ansi>\033\[((?:\d|;)*)([a-zA-Z]))|(?P<text>.)')

    textlen = 0
    has_ansi = False  # Needs RESET code

    for mo in re.finditer(TOKEN_RE, text):

        if textlen >= maxlen:
            break

        kind = mo.lastgroup
        value = mo.group()

        stream.write(value)

        if kind == 'text':
            textlen += len(value)

        if kind == 'ansi':
            has_ansi = True

    if has_ansi:
        # Add ANSI reset sequence
        stream.write('\x1b[0m')

    return stream.getvalue()



class ShopifyObjectBase:

    def __init__(self, data):
        self.data = data

    def __getitem__(self, name):
        return self.data[name]

    def __getattr__(self, name):
        return self.data[name]

    def __repr__(self):
        name = self.__class__.__name__
        return '{}({})'.format(name, repr(self.data))

    @classmethod
    def from_response(cls, data):
        # Allows overriding creation from an API response
        return cls(data)


class Product(ShopifyObjectBase):
    pass



import json
import sys
from collections.abc import Iterable, Mapping

import click
from pygments import highlight
from pygments.formatters import Terminal256Formatter
from pygments.lexers import JsonLexer

from ._types import ShopifyObjectBase


def isatty(stream):
    try:
        return stream.isatty()
    except Exception:
        return False


def print_object(obj, stream=None, fmt=None):
    if stream is None:
        stream = sys.stdout
    _tty = isatty(stream)

    if fmt is None:
        if _tty:
            return print_object_human(obj, stream)
        return print_object_json(obj, stream)

    if fmt == 'json':
        if _tty:
            return print_object_json_pretty(obj, stream)
        return print_object_json(obj, stream)

    if fmt == 'human':
        return print_object_human(obj, stream)

    raise ValueError('Invalid output format: {}'.format(fmt))


def print_object_json(obj, stream):
    json.dump(obj, stream, default=json_default)


def print_object_json_pretty(obj, stream):
    serialized = json.dumps(obj, default=json_default, indent=4)
    highlighted = highlight_json(serialized)
    click.echo(highlighted, file=stream)


def highlight_json(text):
    return highlight(
        text,
        JsonLexer(),
        Terminal256Formatter(style='monokai'))


def json_default(value):

    if isinstance(value, ShopifyObjectBase):
        return value.data

    if isinstance(value, Iterable):
        # Generators etc
        return list(value)

    raise TypeError('Object of type {} is not JSON serializable'
                    .format(type(value)))


def print_object_human(obj, stream):
    # FIXME: actually print objects in a more human-friendly way
    return print_object_json_pretty(obj, stream)
